#!/usr/bin/env python
# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.
# 51 Franklin Street, Fifth Floor
# Boston, MA  02110-1301, USA.

"""
netq - Query data across all nodes in fabric

Usage:
   netq [server <ip-server>|server <text-server>] health (agents|bgp|clag) [around <text-time>] [json]
   netq [server <ip-server>|server <text-server>] health agents [<hostname>] history [json]
   netq [server <ip-server>|server <text-server>] resolve [around <text-time>]
   netq [server <ip-server>|server <text-server>] show info [<ip>|<ip/prefixlen>|<mac>|<hostname>|<text-asn>|<text-vlan>]
   netq [server <ip-server>|server <text-server>] show info [<ip>|<ip/prefixlen>|<mac>|<hostname>|<text-asn>|<text-vlan>] around <text-time>
   netq [server <ip-server>|server <text-server>] show interfaces [<hostname>] [history] [json]
   netq [server <ip-server>|server <text-server>] show interfaces [<hostname>] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show interfaces <hostname> [<remote-interface>] [history] [json]
   netq [server <ip-server>|server <text-server>] show interfaces <hostname> [<remote-interface>] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show lldp [<hostname>] [history]  [json]
   netq [server <ip-server>|server <text-server>] show lldp [<hostname>] [history] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show lldp <hostname> [<remote-interface>] [history] [json]
   netq [server <ip-server>|server <text-server>] show lldp <hostname> [<remote-interface>] [history] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show macs [<hostname>] [<mac>] [vlan <1-4096>] [history]  [json]
   netq [server <ip-server>|server <text-server>] show macs [<hostname>] [<mac>] [vlan <1-4096>]  around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show macs <hostname> [<interface-nexthop>] [<mac>] [vlan <1-4096>] [history]  [json]
   netq [server <ip-server>|server <text-server>] show macs <hostname> [<interface-nexthop>] [<mac>] [vlan <1-4096>]  around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show addresses [<hostname>] [<ip>|<ip/prefixlen>] [history] [json]
   netq [server <ip-server>|server <text-server>] show addresses [<hostname>] [<ip>|<ip/prefixlen>] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show addresses <hostname> [<remote-interface>] [<ip>|<ip/prefixlen>] [<mac>] [history] [json]
   netq [server <ip-server>|server <text-server>] show addresses <hostname> [<remote-interface>] [<ip>|<ip/prefixlen>] [<mac>] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show ip routes [<hostname>] [<ipv4>|<ipv4/prefixlen>] [origin] [json]
   netq [server <ip-server>|server <text-server>] show ip routes [<hostname>] [<ipv4>|<ipv4/prefixlen>] [origin] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show ip routes [<hostname>] [<ipv4/prefixlen>] [origin] [history] [json]
   netq [server <ip-server>|server <text-server>] show ipv6 routes [<hostname>] [<ipv6/prefixlen>] [origin] [history] [json]
   netq [server <ip-server>|server <text-server>] show ipv6 routes [<hostname>] [<ipv6>|<ipv6/prefixlen>] [origin] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show ipv6 routes [<hostname>] [<ipv6>|<ipv6/prefixlen>] [origin] [json]
   netq [server <ip-server>|server <text-server>] show ip neighbors [<hostname>] [<ipv4>] [<mac>] [history] [json]
   netq [server <ip-server>|server <text-server>] show ip neighbors [<hostname>] [<ipv4>] [<mac>] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show ip neighbors <hostname> <remote-interface> [<ipv4>] [<mac>] [history] [json]
   netq [server <ip-server>|server <text-server>] show ip neighbors <hostname> <remote-interface> [<ipv4>] [<mac>] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show ipv6 neighbors [<hostname>] [<ipv6>] [<mac>] [history] [json]
   netq [server <ip-server>|server <text-server>] show ipv6 neighbors [<hostname>] [<ipv6>] [<mac>] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] show ipv6 neighbors <hostname> [<remote-interface>] [<ipv6>] [<mac>] [history] [json]
   netq [server <ip-server>|server <text-server>] show ipv6 neighbors <hostname> [<remote-interface>] [<ipv6>] [<mac>] around <text-time> [json]
   netq [server <ip-server>|server <text-server>] view <hostname> <wildcard-command>
   netq [server <ip-server>|server <text-server>] view around <text-time> <hostname> <wildcard-command>
   netq add server <ip-server>|<text-server>
   netq agent start|stop|status|restart


Options:
   add            : Update configuration
   addresses      : Interface addresses(IP or MAC) across all nodes in fabric
   agent          : Netq agent
   agents         : Check the 'liveness' of the agent across all the nodes
   around         : Go back in time to around ...
   bgp            : Check the status of all the BGP sessions across the fabric
   clag           : Check the status of all the CLAG daemons across the fabric
   health         : Show liveness and status of agent/bgp/mlag in fabric
   help           : Show usage info
   history        : how this infomation has changed with time
   interfaces     : Interfaces across all nodes in fabric
   ip             : IPv4 related info
   ipv6           : IPv6 related info
   json           : provide output in JSON
   info           : Show all relevant info about the provided input
   lldp           : LLDP based neighbor info
   mac            : destination mac address to L2 path trace
   macs           : Show mac address info, with optional VLAN, across the fabric
   neighbor       : IP neighbor cache (ARP/ND) info across the fabric
   neighbors      : IP neighbor info (ARP/ND)
   origin         : owner of route
   resolve        : Annotate input with names and interesting info
   routes         : Show IPv4 route info across the fabric
   show           : Show fabric-wide info
   start          : Start agent
   stop           : Stop agent
   status         : Display agent status
   server         : Specify IP address of DB server
   trace          : Trace path across fabric
   view           : Show output of pre-defined commands on specific node
   vlan           : VLAN
   server         : IP address of DB server
   <1-4096>       : VLAN number, between 1-4096
   <hostname>     : Name of the remote node you want to query
   <interface-nextHop> : The name of the nexthop interface
   <ip-server>    : IP address of DB server
   <ipv4>         : IPv4 address (no mask)
   <ipv6>         : IPv6 address (no mask)
   <ip>           : IPv4 or v6 address (no mask)
   <ipv4/prefixlen> : IPv4 address with mask
   <ipv6/prefixlen> : IPv6 address with mask
   <ip/prefixlen> : IPv4 or IPv6 address with mask
   <mac>          : MAC address
   <remote-interface> : Name of the interface on the remote node
   <text-time>    : Time string such as 10m, 30s, 1h
   <text-server>  : hostname of DB server (must be DNS resolvable)
   <text-asn>     : ASN you're interested in info about
   <text-vlan>    : VLAN you're interested in info about

Help:
    -h --help     Show this screen

"""
from __future__ import absolute_import

import functools
import itertools
import json
import operator
import os
import socket
import subprocess
import sys
from datetime import datetime

import tabulate
from network_docopt import (
    NetworkDocopt,
    get_network_docopt_info
)

from netq.common import utils
from netq.common.enums import (
    BgColors,
    Heartbeat
)
from netq.lib.netq import (
    NetQ,
    NetQException
)

_JSON = []


class _MyTabulate(object):
    @staticmethod
    def __line_row(colsizes, *_):
        return ' '.join('-' * colsize for colsize in colsizes)

    @classmethod
    def __build_row(cls, colsizes, sep, cell_values, _, colaligns):
        values_with_attrs = []
        new_cell_values = []
        for cell_value, colsize, colalign in zip(cell_values,
                                                 colsizes,
                                                 colaligns):
            if colalign == 'right':
                fmt = "{0:>%d}" % colsize
            else:
                fmt = "{0:<%d}" % colsize
            key = fmt.format(cell_value.strip())
            if len(key) > colsize:
                idx = key.rfind(sep, 0, colsize)
                if idx == -1:
                    idx = colsize - 1
                new_cell_values.append(key[idx + 1:])
                key = key[:idx + 1] + ' ' * (colsize - idx - 1)
            else:
                new_cell_values.append('')
            values_with_attrs.append(key)
        line = ' '.join(values_with_attrs).rstrip()
        if any(new_cell_values):
            return line + '\n' + cls.__build_row(colsizes, sep,
                                                 new_cell_values, _,
                                                 colaligns)
        else:
            return line

    @classmethod
    def paginate(cls, data, headers, colsizes, sep=',', first=False):
        colsizes = [max(x, len(y)) for x, y in zip(colsizes, headers)]
        linebelowheader = functools.partial(cls.__line_row, colsizes)
        headerrow = datarow = functools.partial(cls.__build_row, colsizes, sep)
        simple = tabulate.TableFormat(lineabove=None,
                                      linebelowheader=linebelowheader,
                                      linebetweenrows=None,
                                      linebelow=None,
                                      headerrow=headerrow,
                                      datarow=datarow,
                                      padding=0,
                                      with_header_hide=['lineabove',
                                                        'linebelow'])
        if first:
            return tabulate.tabulate(data, headers=headers, tablefmt=simple)
        else:
            return tabulate.tabulate(data, headers=[], tablefmt=simple)


def _tabulate(output, headers=None, jsonify=False, first=True, colsizes=None):
    if jsonify:
        out = []
        sanitized_headers = []
        for header in headers:
            sanitized_headers.append(
                ''.join(i.lower() if idx == 0 else i.lower().title()
                        for idx, i in enumerate(header.split(' ')))
            )
        for ele in output:
            out.append(dict((key, utils.strip_colors(item))
                            for key, item in zip(sanitized_headers, ele)))
        _JSON.append(out)
    else:
        if colsizes is not None:
            print _MyTabulate.paginate(output, headers=headers,
                                       colsizes=colsizes, first=first)
        else:
            output = tabulate.tabulate(output, headers)
            if first:
                print output
            else:
                print output[output.find('\n', output.find('\n') + 1) + 1:]


def show_info(cli):
    """
    """
    mac = cli.get('<mac>') or '*'
    hostname = cli.get('<hostname>') or '*'
    asn = cli.get('<asn>') or '*'
    vlan = cli.get('<vlan>') or '*'
    jsonify = cli.get('json')
    iface = '*'
    oif = '*'
    ip_address = cli.get('<ip>') or cli.get('<ip/prefixlen>') or '*'

    if ip_address is not '*':
        is_ipv6 = ':' in ip_address
        first = True
        for out in cli.netq.show_addresses(hostname, iface, ip_address, mac):
            if out and not jsonify and first:
                print 'Interface Address records are'
            _tabulate(sorted(out),
                      ['IP', 'Node', 'Interface', 'Update Time'],
                      first=first,
                      jsonify=jsonify),
            first = False

        first = True
        for out in cli.netq.show_routes(ip_address, hostname, is_ipv6, False):
            if out and not jsonify and first:
                print 'Route info about prefix %s in fabric' % ip_address
            _tabulate(sorted(out),
                      ['IP', 'Node', 'Nexthops', 'Origin',
                       'Update Time'],
                      first=first,
                      jsonify=jsonify),
            first = False

        out = cli.netq.show_router_id(ip_address)
        if out:
            if not jsonify:
                print 'List of nodes with Router ID %s' % ip_address

            _tabulate(sorted(out),
                      ['Router ID', 'Node', 'ASN', 'Protocol'], jsonify)

    elif mac is not '*':
        first = True
        for out in cli.netq.show_macvlan(mac, vlan, hostname, oif):
            if out and not jsonify and first:
                print 'List of nodes containing MAC %s' % mac
            _tabulate(sorted(out),
                      ['MAC', 'VLAN', 'Node Name', 'Egress Port',
                       'Update Time'],
                      jsonify=jsonify,
                      first=first)
            first = False

        print cli.netq.show_info(ip_address, mac, hostname)

    elif asn is not '*':
        asn = int(asn)
        out = cli.netq.show_asn(asn)
        if out:
            if jsonify:
                print 'List of nodes with ASN %s' % asn

            _tabulate(sorted(out),
                      headers=['ASN', 'Node', 'Router ID'],
                      jsonify=jsonify)

    elif vlan is not '*':
        _ = int(vlan)

    elif hostname is not '*':
        try:
            print cli.netq.show_info(ip_address, mac, hostname)
        except NetQException as ex:
            print str(ex)


def _run(cli):
    """Once upon a time...
    """
    server_addr = None
    if cli.get('server'):
        server_addr = cli.get('<ip-server>')
        if server_addr is None:
            server_name = cli.get('<text-server>')
            try:
                server_addr = socket.getaddrinfo(server_name, 0)[0][4][0]
            except socket.gaierror:
                print "Server %s is not known.Exiting" % server_name
                sys.exit(1)

    start_time = None
    if cli.get('around'):
        start_time = cli.get('<text-time>')

    cli.netq = NetQ(server_addr, start_time)
    netq = cli.netq

    jsonify = cli.get('json')
    history = cli.get('history')

    # TESTED
    if cli.get('add'):
        if os.getuid() != 0:
            raise RuntimeError(utils.color_wrap(
                BgColors.RED, 'Please run this command as root'))
        print netq.update_config(server_addr)

    elif cli.get('agent'):
        if cli.get('start'):
            if os.getuid() != 0:
                raise RuntimeError('Please run this command as root')
            try:
                subprocess.check_output(
                    'systemctl enable netq-agent > /dev/null 2>&1',
                    shell=True
                )
                subprocess.check_output(
                    'systemctl reset-failed netq-agent'.split()
                )
                subprocess.check_output(
                    'systemctl start netq-agent'.split()
                )
                print utils.color_wrap(BgColors.GREEN, 'Success!')
            except subprocess.CalledProcessError:
                raise RuntimeError(utils.color_wrap(BgColors.RED,
                                                    'Failed to start agent'))
        elif cli.get('stop'):
            if os.getuid() != 0:
                raise RuntimeError('Please run this command as root')
            try:
                subprocess.check_output(
                    'systemctl stop netq-agent'.split()
                )
                subprocess.check_output(
                    'systemctl disable netq-agent > /dev/null 2>&1',
                    shell=True
                )
                print utils.color_wrap(BgColors.GREEN, 'Success!')
            except subprocess.CalledProcessError:
                raise RuntimeError(utils.color_wrap(BgColors.RED,
                                                    'Failed to stop agent'))
        elif cli.get('status'):
            try:
                subprocess.check_output('systemctl status netq-agent'.split())
                print utils.color_wrap(BgColors.GREEN, 'Running...')
            except subprocess.CalledProcessError as ex:
                print utils.color_wrap(BgColors.RED, 'Stopped...')
                return ex.returncode
        elif cli.get('restart'):
            if os.getuid() != 0:
                raise RuntimeError('Please run this command as root')
            try:
                subprocess.check_output(
                    'systemctl enable netq-agent > /dev/null 2>&1',
                    shell=True
                )
                subprocess.check_output(
                    'systemctl reset-failed netq-agent'.split()
                )
                subprocess.check_output(
                    'systemctl restart netq-agent'.split()
                )
                print utils.color_wrap(BgColors.GREEN, 'Success!')
            except subprocess.CalledProcessError:
                raise RuntimeError(utils.color_wrap(BgColors.RED,
                                                    'Failed to restart agent'))

    elif cli.get('health'):
        if cli.get('agents'):
            if history:
                hostname = cli.get('<hostname>') or '*'
                out = []
                restarts = 0
                for key, grouper in itertools.groupby(
                        netq.show_agent_history(hostname=hostname),
                        key=operator.itemgetter(0)
                ):
                    items = sorted(grouper, key=operator.itemgetter(1))
                    if len(items) > 1:
                        restarts = len(items)
                        color = BgColors.RED
                    else:
                        restarts = 0
                        color = BgColors.GREEN

                    last_beat = (datetime.utcnow() - items[-1][2])
                    if last_beat.total_seconds() > Heartbeat.HEARTBEAT_DEATH:
                        cur_color = BgColors.RED
                    elif last_beat.total_seconds() <= Heartbeat.LIVELINESS_RATE:
                        cur_color = BgColors.GREEN
                    else:
                        cur_color = BgColors.YELLOW

                    if hostname != '*':
                        # Provide detailed output for this scenario
                        for item in items:
                            out.append([key, str(item[1])])
                    else:
                        out.append([utils.color_wrap(color, key),
                                    str(items[0][1]),
                                    utils.color_wrap(cur_color,
                                                     str(items[-1][1])),
                                    utils.color_wrap(color, '%d' % restarts)])

                if hostname == '*':
                    _tabulate(out,
                              headers=['Node', 'First Connect time',
                                       'Last Connect Time',
                                       'NumRestarts'], jsonify=jsonify)
                else:
                    if not jsonify and restarts:
                        print utils.color_wrap(BgColors.RED,
                                               'Number of Restarts: %d' % \
                                               restarts)
                    _tabulate(out, headers=['Node', 'Connect Time'],
                              jsonify=jsonify)


            else:
                out = netq.show_status()
                _tabulate(sorted(out),
                          headers=['Node Name', 'Connect Time',
                                   'Time Since Connect', 'Status'],
                          jsonify=jsonify)

        elif cli.get('bgp'):
            out = netq.show_bgp_status()
            if out:
                timestr = out[-1]
                if not jsonify:
                    print timestr

                _tabulate(out[:-1],
                          headers=['Node', 'Peer', 'Reason'],
                          jsonify=jsonify)
            else:
                if jsonify:
                    print '[{}]'
                else:
                    print utils.color_wrap(BgColors.GREEN,
                                           'No BGP sessions in failed state')

        elif cli.get('clag'):
            failed_node_lst = netq.show_clag_status()

            if failed_node_lst == []:
                if not jsonify:
                    print utils.color_wrap(BgColors.GREEN,
                                           'No CLAG sessions in failed state')
                else:
                    print '[{}]'
            else:
                print _tabulate(failed_node_lst,
                                headers=['Failed Node', 'Reason'],
                                jsonify=jsonify,
                                colsizes=[20, 60])

    elif cli.get('resolve'):
        if not sys.stdin.isatty():
            print netq.resolve_output()

    elif cli.get('trace'):
        if cli.get('mac'):
            netq.path_trace_mac(cli.get('mac'),
                                cli.get('vlan'),
                                cli.get('<hostname-from>'))

    elif cli.get('view'):
        hostname = cli.get('<hostname>')
        key = cli.get('<wildcard-command>') or ''
        args = key.split()
        try:
            print netq.view(hostname, args)
        except NetQException as ex:
            print str(ex)

    elif cli.get('show'):
        if cli.get('info'):
            show_info(cli)

        elif cli.get('interfaces'):
            hostname = cli.get('<hostname>') or '*'
            interface = cli.get('<remote-interface>') or '*'
            headers = ['Node', 'Interface', 'Admin State', 'Carrier',
                       'Type', 'LLDP', 'MTU', 'Present', 'Update Time']
            if history:
                out = netq.show_link_history(hostname, interface)
                if out:
                    _tabulate(out, headers, jsonify=jsonify)
                else:
                    print 'No matching Link found'
            else:
                first = True
                for out in netq.show_links(hostname, interface):
                    _tabulate(sorted(out), headers, jsonify=jsonify,
                              first=first)
                    first = False

        elif cli.get('lldp'):
            node = cli.get('<hostname>') or '*'
            iface = cli.get('<remote-interface>') or '*'
            out = netq.show_neighbors(node, iface)
            if out:
                if not jsonify:
                    if node is not None:
                        if iface is not None:
                            print 'LLDP peer info for %s:%s' % (node, iface)
                        else:
                            print 'LLDP Peer Info for %s' % node
                    else:
                        print 'LLDP Info for fabric'
                _tabulate(sorted(out),
                          headers=['Node', 'Interface', 'LLDP Peer',
                                   'Peer Int', 'Update Time'],
                          jsonify=jsonify)
            else:
                if not jsonify:
                    print 'No LLDP record found'
                else:
                    print '{}'

        elif cli.get('macs'):
            mac = cli.get('<mac>') or '*'
            vlan = cli.get('<1-4096>') or '*'
            if vlan is not '*':
                vlan = int(vlan)

            hostname = cli.get('<hostname>') or '*'
            oif = cli.get('<interface-nexthop>') or '*'

            if history:
                out = netq.show_mac_history(mac, vlan, hostname, oif)
                if out:
                    _tabulate(out,
                              ['MAC', 'VLAN', 'Node Name', 'Egress Port',
                               'Present', 'Update Time'],
                              jsonify=jsonify)
                else:
                    print 'No matching MAC entry found'
            else:
                first = True
                for out in netq.show_macvlan(mac, vlan, hostname, oif):
                    if not jsonify and first:
                        if vlan != '*':
                            print (
                                'List of nodes containing MAC %s in VLAN %s' %
                                (mac, vlan)
                            )
                        else:
                            print 'List of nodes containing MAC %s' % mac
                    _tabulate(sorted(out),
                              headers=['MAC', 'VLAN', 'Node Name',
                                       'Egress Port', 'Update Time'],
                              jsonify=jsonify,
                              first=first)
                    first = False
                if first:
                    if not jsonify:
                        print 'No matching MAC entry found'
                    else:
                        print '{}'

        elif cli.get('addresses'):
            node = cli.get('<hostname>') or '*'
            iface = cli.get('<remote-interface>') or '*'
            ip_address = cli.get('<ip>') or cli.get('<ip/prefixlen>') or '*'
            mac = cli.get('<mac>') or '*'

            if history:
                out = netq.show_address_history(node, iface, ip_address, mac)
                if out:
                    _tabulate(out,
                              ['IP', 'Node', 'Interface', 'Present',
                               'Update Time'],
                              jsonify=jsonify)
                else:
                    print 'No matching Address found'
            else:
                first = True
                for out in netq.show_addresses(node, iface, ip_address, mac):
                    if not jsonify and first:
                        print 'Interface Address records are'
                    _tabulate(sorted(out),
                              ['IP', 'Node', 'Interface', 'Update Time'],
                              jsonify=jsonify,
                              first=first)
                    first = False
                if first:
                    if not jsonify:
                        print 'No matching Interface Address entry found'
                    else:
                        print '{}'

        elif cli.get('routes'):
            hostname = cli.get('<hostname>') or '*'
            if cli.get('ip'):
                ip_address = (
                    cli.get('<ipv4>') or cli.get('<ipv4/prefixlen>') or '*'
                )
                is_ipv6 = False
            else:
                ip_address = (
                    cli.get('<ipv6>') or cli.get('<ipv6/prefixlen>') or '*'
                )
                is_ipv6 = True
            if cli.get('origin'):
                origin = True
            else:
                origin = '*'

            if history:
                out = netq.show_route_history(ip_address,
                                              hostname,
                                              ipv6=is_ipv6)
                if out:
                    _tabulate(out,
                              ['IP', 'Node', 'Nexthops', 'Origin',
                               'Present', 'Update Time'],
                              jsonify=jsonify)
            else:
                first = True
                for out in netq.show_routes(ip_address, hostname, is_ipv6,
                                            origin):
                    if not jsonify and first:
                        if ip_address:
                            if hostname:
                                print ('Route info about prefix %s on '
                                       'host %s' % (ip_address, hostname))
                            else:
                                print ('Route info about prefix %s in fabric'
                                       % ip_address)
                        elif hostname:
                            print 'Routes in host %s FIB' % hostname
                        else:
                            print 'List of all routes in fabric'
                    # Print routes with origin first
                    _tabulate(sorted(out),
                              ['Origin', 'Table', 'IP', 'Node',
                               'Nexthops', 'Update Time'],
                              jsonify=jsonify,
                              first=first,
                              colsizes=[3, 3, 16, 16, 25, 16])
                    first = False
                if first:
                    if not jsonify:
                        print 'No matching routes found'
                    else:
                        print '{}'

        elif cli.get('neighbors'):
            if cli.get('ip'):
                ip_address = cli.get('<ipv4>') or '*'
                is_ipv6 = False
            else:
                ip_address = cli.get('<ipv6>') or '*'
                is_ipv6 = True

            mac = cli.get('<mac>') or '*'
            hostname = cli.get('<hostname>') or '*'
            iface = cli.get('<remote-interface>') or '*'
            if history:
                out = netq.show_ip_neighbor_history(hostname=hostname,
                                                    ip_address=ip_address,
                                                    ifname=iface,
                                                    mac=mac,
                                                    ipv6=is_ipv6)
                if out:
                    _tabulate(out,
                              ['IP', 'Node', 'Interface', 'MAC',
                               'Present', 'Update Time'],
                              jsonify)
                else:
                    print 'No IP neighbor entry found'
            else:
                first = True
                for out in netq.show_ip_neighbors(ip_address, hostname,
                                                  iface, mac, is_ipv6):
                    if not jsonify and first:
                        if ip_address is not '*':
                            print 'IP Neighbor info for %s' % ip_address
                        else:
                            print 'IP Neighbor Info for fabric'
                    _tabulate(sorted(out),
                              ['IP', 'Node', 'Interface', 'MAC',
                               'Update Time'],
                              jsonify=jsonify,
                              first=first)
                    first = False
                if first:
                    print 'No IP neighbor entry found'


def main():
    (print_options, ended_with_space, sys.argv) = (
        get_network_docopt_info(sys.argv)
    )
    cli = NetworkDocopt(__doc__)

    # Add custom completers
    cli.add_tag_matcher('<hostname>', NetQ.hostname_match, True)
    cli.add_tab_complete_hook('<hostname>', NetQ.hostname_complete)
    cli.add_tag_matcher('<wildcard-command>', NetQ.view_match, True)
    cli.add_tab_complete_hook('<wildcard-command>', NetQ.view_complete)
    cli.add_tag_matcher('<remote-interface>', NetQ.interface_match, True)
    cli.add_tab_complete_hook('<remote-interface>', NetQ.interface_complete)
    cli.add_tag_matcher('<interface-nexthop>', NetQ.interface_match, True)

    try:
        if print_options:
            cli.print_options(ended_with_space)
        else:
            cli.run(argv=[sys.argv[0]] + (sys.argv[1:] or ['health',
                                                           'agents']))
            out = _run(cli)
            if _JSON:
                print json.dumps(_JSON)
            return out
    except IOError as e:
        # stdout is closed, no point in continuing
        # Attempt to close them explicitly to prevent cleanup problems:
        if str(e).lower() == 'broken pipe':
            exit(0)
        try:
            sys.stdout.close()
        except IOError:
            pass
        try:
            sys.stderr.close()
        except IOError:
            pass
