# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.
# 51 Franklin Street, Fifth Floor
# Boston, MA  02110-1301, USA.


class SortedSet(object):

    def __init__(self, score, value):
        self.__left = None
        self.__right = None
        self.__score = float(score)
        self.__value = value

    def __inorder(self):
        if self.__left is not None:
            for node in self.__left.__inorder():
                yield node
        yield self
        if self.__right is not None:
            for node in self.__right.__inorder():
                yield node

    def __insert(self, score, value):
        if self.__score < score:
            if self.__right is not None:
                self.__right.__insert(score, value)
            else:
                self.__right = SortedSet(score, value)
        elif self.__score > score:
            if self.__left is not None:
                self.__left.__insert(score, value)
            else:
                self.__left = SortedSet(score, value)

    def __is_unique(self, value):
        return all(value != node.__value for node in self.__inorder())

    def __range(self, func, start, end, score_cast_func, withscores):
        nodes = [node for node in self.__inorder()]
        if start is None:
            start = 0
        if end is None:
            end = len(nodes)
        for node in func(nodes[start:end]):
            if withscores:
                yield node.__value, score_cast_func(node.__score)
            else:
                yield node.__value

    def __rangebyscore(self, func, start, end, score_cast_func, withscores):
        nodes = [node for node in self.__inorder()]
        for node in func(nodes):
            if start <= node.__score <= end:
                if withscores:
                    yield node.__value, score_cast_func(node.__score)
                else:
                    yield node.__value

    def insert(self, score, value):
        if self.__is_unique(value):
            self.__insert(score, value)

    def latest(self, score_cast_func=float, withscores=False):
        right = self
        while right is not None:
            right = right.__right
        if withscores:
            return right.__value, score_cast_func(right.__score)
        else:
            return right.__value

    def range(self, start=None, end=None, score_cast_func=float,
              withscores=False):
        for val in self.__range(iter, start, end, score_cast_func,
                                withscores):
            yield val

    def revrange(self, start=None, end=None, score_cast_func=float,
                 withscores=False):
        for val in self.__range(reversed, start, end, score_cast_func,
                                withscores):
            yield val

    def rangebyscore(self, start=None, end=None, score_cast_func=float,
                     withscores=True):
        for val in self.__rangebyscore(iter, start, end, score_cast_func,
                                       withscores):
            yield val

    def revrangebyscore(self, start=None, end=None, score_cast_func=float,
                        withscores=True):
        for val in self.__rangebyscore(reversed, end, start, score_cast_func,
                                       withscores):
            yield val




