# vim: tabstop=4 shiftwidth=4 softtabstop=4
# Copyright 2016 Cumulus Networks, Inc. All rights reserved.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.
# 51 Franklin Street, Fifth Floor
# Boston, MA  02110-1301, USA.
import time

from netq.orm import objects, HOSTNAME
from netq.orm.memdb.db import SortedSet
from netq.orm.memdb.query import MemQuery


class _MemModel(objects.Model):
    query_cls = MemQuery
    hostname = objects.String(json_key='hostname')
    timestamp = objects.Float(json_key='timestamp')

    def __init__(self):
        self.hostname = HOSTNAME
        self.timestamp = time.time()
        super(_MemModel, self).__init__()

    def save(self, key, score, value):
        if key in self.query.map:
            self.query.map[key].insert(score, value)
        else:
            self.query.map[key] = SortedSet(score, value)


class MemNode(_MemModel):
    bgp_peer_count = objects.Integer(json_key='bgpPeerCount')
    bgp_active_peer_count = objects.Integer(json_key='bgpActivePeerCount')
    bgp_rib_entries = objects.Integer(json_key='bgpRibEntries')
    bgp_rib_mem = objects.Integer(json_key='bgpRibMem')
    bgp_asn = objects.Integer(json_key='bgpAsn')
    bgp_router_id = objects.String(json_key='bgpRouterId')

    clag_role = objects.String(json_key='clagRole')
    clag_state = objects.Boolean(json_key='clagState')
    clag_peer_role = objects.String(json_key='clagPeerRole')
    clag_peer_state = objects.Boolean(json_key='clagPeerState')
    clag_peer_if = objects.String(json_key='clagPeerIf')
    clag_sysmac = objects.String(json_key='sysMac')
    clag_backup_ip = objects.String(json_key='backupIp')
    clag_singly_attached_bonds = objects.Json(json_key='singlyAttachedBonds')
    clag_conflicted_bonds = objects.Json(json_key='conflictedBonds')
    clag_proto_down_bonds = objects.Json(json_key='protoDownBonds')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.bgp_peer_count,
                cls.bgp_active_peer_count, cls.bgp_rib_entries,
                cls.bgp_rib_mem, cls.bgp_asn, cls.bgp_router_id,
                cls.clag_state, cls.clag_peer_role, cls.clag_role,
                cls.clag_peer_if, cls.clag_sysmac, cls.clag_backup_ip,
                cls.clag_peer_state, cls.clag_singly_attached_bonds,
                cls.clag_proto_down_bonds, cls.clag_conflicted_bonds)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return None

    def save(self, **kwargs):
        super(MemNode, self).save(MemNode.obj_to_key(self.__dict__),
                                  MemNode.obj_to_score(self.__dict__),
                                  True)


class BgpSession(_MemModel):
    peer_state = objects.String(json_key='peerState')
    peer_name = objects.String(json_key='peerName')
    peer_asn = objects.Integer(json_key='peerAsn')
    reason = objects.String(json_key='reason')
    pfx_rcv_count = objects.Integer(json_key='prefixRcvCount')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.peer_state, cls.peer_name,
                cls.peer_asn, cls.reason)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return None

    def save(self, **kwargs):
        super(BgpSession, self).save(BgpSession.obj_to_key(self.__dict__),
                                     BgpSession.obj_to_score(self.__dict__),
                                     True)


class Lldp(_MemModel):
    local_intf = objects.String(json_key='localIntf')
    lldp_peer = objects.String(json_key='lldpPeer')
    lldp_intf = objects.String(json_key='lldpIntf')
    lldp_peer_bridge = objects.Boolean(json_key='lldpPeerBridge')
    lldp_peer_router = objects.Boolean(json_key='lldpPeerRouter')
    lldp_peer_os = objects.String(json_key='lldpPeerOS')

    @classmethod
    def key_fmt(cls, key_type=None):
        return (cls.__name__, cls.hostname, cls.local_intf, cls.lldp_peer,
                cls.lldp_intf, cls.lldp_peer_bridge, cls.lldp_peer_router,
                cls.lldp_peer_os)

    @classmethod
    def score_fmt(cls, key_type=None):
        return cls.timestamp

    @classmethod
    def val_fmt(cls):
        return None

    def save(self, **kwargs):
        super(Lldp, self).save(Lldp.obj_to_key(self.__dict__),
                               Lldp.obj_to_score(self.__dict__),
                               True)
